#include <iostream>

#include "Rectangle/ContourAnalyzer.h"
#include "webcam/Webcam.h"
#include "UI.h"
#include "Rectangle/Confidence.h"
#include "Rectangle/BoundTracker.h"
#include "Rectangle/Ruler.h"
#include <memory>

enum State {
	detection, track, cross_check
};

State state = detection;
Webcam webcam;
std::shared_ptr<int> lastKey {new int(0)};
std::shared_ptr<UI> ui {new UI(lastKey)};
ContourAnalyzer ca { ui, lastKey };
Confidence conf { lastKey, ui, 5, 0.74};
BoundTracker tracker { lastKey, ui };
cv::Point statusTextPos {5, 15};
cv::Scalar textFontColor { 20, 20, 20};
cv::Scalar rectangleBoundColor { 255, 0, 0 };
Ruler ruler{ui, lastKey};

unsigned frameCount = 0;
cv::Point center;
void run() {
	switch (state) {
	case detection:
		cv::putText(*(ui->drawnFrame()), "Looking for rectangle...",
				statusTextPos, cv::FONT_HERSHEY_PLAIN, 0.8, textFontColor);
		ca.analyze();
		ca.convertContoursToBounds();
		if (conf.check(ca.getBounds())) {
			cv::Rect largest;
			std::vector<cv::Rect> rects = conf.rectangleBounds();
			for (unsigned i = 0; i < rects.size(); i++) {
				if (rects[i].area() > largest.area()) {
					largest = rects[i];
				}
			}
			tracker.init(largest);
			state = track;
		}
		break;

	case track:
	{
		int growBy = 30;
		cv::Rect scaledSearchArea;
		cv::putText(*(ui->drawnFrame()), "Tracking...", statusTextPos, cv::FONT_HERSHEY_PLAIN, 0.8, textFontColor);
		if (!tracker.track()) {
			state = cross_check;
		}
		scaledSearchArea = tracker.getROI();
		scaledSearchArea += cv::Size{growBy*2, growBy*2};
		scaledSearchArea.x -= growBy;
		scaledSearchArea.y -= growBy;

		scaledSearchArea.x = std::max(scaledSearchArea.x, 0);
		scaledSearchArea.y = std::max(scaledSearchArea.y, 0);

		scaledSearchArea.x = std::min(scaledSearchArea.x, ui->getWidth()-scaledSearchArea.width -1);
		scaledSearchArea.y = std::min(scaledSearchArea.y, ui->getHeight()-scaledSearchArea.height -1);

		ca.analyze(scaledSearchArea);

		ca.convertContoursToBounds();

		if (ca.getBounds().size() > 0) {
			cv::Rect largest;
			for (unsigned i = 0; i < ca.getBounds().size(); i++) {
				if (ca.getBounds()[i].area() > largest.area()) {
					largest = ca.getBounds()[i];
				}
			}

			ruler.setDetectedWidth(largest.width);

			largest.x += scaledSearchArea.x;
			largest.y += scaledSearchArea.y;
			tracker.init(largest);
		}

		cv::Rect2d bound = tracker.getROI();
		center.x = bound.width/2;
		center.y = bound.height/2;
		center.x += bound.x;
		center.y += bound.y;

		cv::circle(*(ui->drawnFrame()), center, 3, rectangleBoundColor, 2);

	}
	break;

	case cross_check:
		cv::putText(*(ui->drawnFrame()), "Cross track...", statusTextPos,
				cv::FONT_HERSHEY_PLAIN, 0.8, textFontColor);
		ca.analyze();
		ca.convertContoursToBounds();
		tracker.track();

		std::vector<cv::Rect> rects = ca.getBounds();
		cv::Rect2d largest;
		if (rects.size() > 0) {
			for (unsigned i = 0; i < rects.size(); i++) {
				if (rects[i].area() > largest.area()) {
					largest = rects[i];
				}
			}

			unsigned largestArea = tracker.getROI().area() > largest.area() ? tracker.getROI().area() : largest.area();

			if ((tracker.getROI() | largest).area() <= largestArea * (1.25)) {
				tracker.init(largest);
				state = track;
				frameCount = 0;
			} else {
				frameCount++;
				if (frameCount > 20) {
					state = detection;
					frameCount = 0;
				}
			}
			cv::rectangle(*(ui->drawnFrame()), largest, cv::Scalar { 100, 0, 0});
		} else {
			frameCount++;
			if (frameCount > 30) {
				state = detection;
				frameCount = 0;
			}
		}

		cv::Rect2d bound = tracker.getROI();
		center.x = bound.width/2;
		center.y = bound.height/2;
		center.x += bound.x;
		center.y += bound.y;

		cv::circle(*(ui->drawnFrame()), center, 3, rectangleBoundColor);
		break;
	}
	ruler.update();
	ui->render();
}

int main(int argc, char** argv) {
	while (true) {
		webcam.update();
		cv::Mat frame = webcam.getFrame();
		ui->setOriginalFrame(frame);

		*lastKey = cv::waitKey(33);
		if (*lastKey > 0) {
			std::cout << "key pressed: " << *lastKey << std::endl;
			if (*lastKey == 27) {
				break;
			}
		}
		run();
		*lastKey = -1;
	}

	cv::destroyAllWindows();
	return 0;
}
